import copy
import os
import signal
from textwrap import dedent

import matplotlib
matplotlib.use('Qt5Agg')

import yaml
from PyQt5.QtCore import QTimer
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QMainWindow
from PyQt5.QtWidgets import QMessageBox

from mdt.gui.maps_visualizer.actions import NewConfigAction, SetMapsToShow, NewDataAction
from mdt.gui.maps_visualizer.config_tabs.tab_general import TabGeneral
from mdt.gui.maps_visualizer.config_tabs.tab_map_specific import TabMapSpecific
from mdt.gui.maps_visualizer.config_tabs.tab_textual import TabTextual
from mdt.gui.maps_visualizer.design.ui_save_image_dialog import Ui_SaveImageDialog
from mdt.lib.nifti import is_nifti_file

import mdt
from mdt.gui.maps_visualizer.base import DataConfigModel, \
    QtController
from mdt.gui.maps_visualizer.renderers.base import PlottingFrameInfoViewer
from mdt.visualization.maps.base import DataInfo, SimpleDataInfo, MapPlotConfig
from mdt.visualization.maps.utils import get_shortest_unique_names
from mdt.gui.maps_visualizer.renderers.matplotlib_renderer import MatplotlibPlotting
from mdt.gui.model_fit.design.ui_about_dialog import Ui_AboutDialog
from mdt.gui.utils import center_window, QtManager, get_script_file_header_text, image_files_filters, \
    enable_pyqt_exception_hook
from mdt.gui.maps_visualizer.design.ui_MainWindow import Ui_MapsVisualizer


class MapsVisualizerWindow(QMainWindow, Ui_MapsVisualizer):

    def __init__(self, controller, parent=None):
        """Instantiate the maps GUI

        Args:
            controller (mdt.gui.maps_visualizer.base.Controller): the controller to use for updating the views
        """
        super().__init__(parent)
        self.setupUi(self)

        self._controller = controller
        self._controller.model_updated.connect(self.update_model)

        self.setAcceptDrops(True)

        self._coordinates_label = QLabel()

        self.statusBar().addPermanentWidget(self._coordinates_label)
        self.statusBar().setStyleSheet("QStatusBar::item { border: 0px solid black }; ")

        self.plotting_info_to_statusbar = PlottingFrameInfoToStatusBar(self._controller, self._coordinates_label)
        self.plotting_frame = MatplotlibPlotting(controller, parent=parent,
                                                 plotting_info_viewer=self.plotting_info_to_statusbar)
        self.plotLayout.addWidget(self.plotting_frame)

        self.tab_general = TabGeneral(controller, self)
        self.generalTabPosition.addWidget(self.tab_general)

        self.tab_specific = TabMapSpecific(controller, self)
        self.mapSpecificTabPosition.addWidget(self.tab_specific)

        self.tab_textual = TabTextual(controller, self)
        self.textInfoTabPosition.addWidget(self.tab_textual)

        self.auto_rendering.setChecked(True)
        self.auto_rendering.stateChanged.connect(self._set_auto_rendering)
        self.manual_render.clicked.connect(lambda: self.plotting_frame.redraw())

        self.actionNew_window.triggered.connect(lambda: start_gui(app_exec=False))
        self.actionAbout.triggered.connect(lambda: AboutDialog(self).exec_())
        self.actionAdd_new_files.triggered.connect(self._add_new_files)
        self.action_Clear.triggered.connect(self._remove_files)
        self.actionSaveImage.triggered.connect(lambda: ExportImageDialog(self, self.plotting_frame,
                                                                         self._controller).exec_())
        self.actionSave_settings.triggered.connect(lambda: self._save_settings())
        self.actionLoad_settings.triggered.connect(lambda: self._load_settings())

        self.undo_config.setDisabled(not self._controller.has_undo())
        self.redo_config.setDisabled(not self._controller.has_redo())

        self.undo_config.clicked.connect(lambda: self._controller.undo())
        self.redo_config.clicked.connect(lambda: self._controller.redo())

        self._qdialog_basedir_set = False

    @pyqtSlot(DataConfigModel)
    def update_model(self, model):
        self.undo_config.setDisabled(not self._controller.has_undo())
        self.redo_config.setDisabled(not self._controller.has_redo())
        self._set_qdialog_basedir()

    def _set_qdialog_basedir(self):
        if not self._qdialog_basedir_set:
            data = self._controller.get_model().get_data()
            for map_name, file_path in data.get_file_paths().items():
                if file_path:
                    QFileDialog().setDirectory(file_path)
                    self._qdialog_basedir_set = True
                    return

    def resizeEvent(self, event):
        ExportImageDialog.plot_frame_resized()

    def _add_new_files(self):
        new_files = QFileDialog(self).getOpenFileNames(caption='Nifti files',
                                                       filter=';;'.join(image_files_filters))
        if new_files[0]:
            self._add_new_maps(new_files[0])

    def _add_new_maps(self, paths):
        """Add the given set of file paths to the current visualization.

        This looks at the current set of file paths (from ``self._controller.get_config().get_data()``)
        and the given new set of file paths and creates a new merged dataset with all files.

        Since it is possible that adding new maps leads to naming collisions this function can rename both the
        old and the new maps to better reflect the map names.

        Args:
            paths (list of str): the list of file paths to add to the visualization

        Returns:
            list of str: the display names of the newly added maps
        """
        def get_file_paths(data_info):
            """Get the file paths"""
            paths = []
            for map_name in data_info.get_map_names():
                file_path = data_info.get_file_path(map_name)
                if file_path:
                    paths.append(file_path)
                else:
                    paths.append(map_name)
            return paths

        def get_changes(old_data_info, new_data_info):
            additions = {}
            removals = []
            name_updates = {}

            paths = get_file_paths(old_data_info) + get_file_paths(new_data_info)
            unique_names = get_shortest_unique_names(paths)
            new_map_names = unique_names[len(old_data_info.get_map_names()):]

            for old_name, new_name in zip(old_data_info.get_map_names(), unique_names):
                if old_name != new_name:
                    removals.append(old_name)
                    additions[new_name] = old_data_info.get_single_map_info(old_name)
                    name_updates[old_name] = new_name

            for new_map_name, new_map_rename in zip(new_data_info.get_map_names(), new_map_names):
                additions[new_map_rename] = new_data_info.get_single_map_info(new_map_name)

            return additions, removals, name_updates, new_map_names

        current_model = self._controller.get_model()
        current_data = current_model.get_data()

        adds, rems, name_updates, new_map_names = get_changes(current_data, SimpleDataInfo.from_paths(paths))
        data = current_data.get_updated(adds, removals=rems)
        config = copy.deepcopy(current_model.get_config())

        new_maps_to_show = []
        for map_name in config.maps_to_show:
            if map_name in name_updates:
                new_maps_to_show.append(name_updates[map_name])
            else:
                new_maps_to_show.append(map_name)
        config.maps_to_show = new_maps_to_show

        new_map_plot_options = {}
        for map_name, plot_options in config.map_plot_options.items():
            if map_name in name_updates:
                new_map_plot_options[name_updates[map_name]] = plot_options
            else:
                new_map_plot_options[map_name] = plot_options
        config.map_plot_options = new_map_plot_options

        if not len(current_data.get_map_names()):
            config.slice_index = data.get_max_slice_index(config.dimension) // 2

        self._controller.apply_action(NewDataAction(data, config=config))
        return new_map_names

    def _remove_files(self):
        data = SimpleDataInfo({})
        config = MapPlotConfig()
        self._controller.apply_action(NewDataAction(data, config))

    @pyqtSlot()
    def _set_auto_rendering(self):
        auto_render = self.auto_rendering.isChecked()
        self.plotting_frame.set_auto_rendering(auto_render)
        if auto_render:
            self.plotting_frame.redraw()

    def send_sigint(self, *args):
        self.close()

    def dragEnterEvent(self, event):
        """Function to allow dragging nifti files in the viewer for viewing purpose."""
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """One or more files where dropped in the GUI, load all the nifti files among them."""
        nifti_paths = []
        for url in event.mimeData().urls():
            path = url.toLocalFile()
            if os.path.isfile(path) and is_nifti_file(path):
                nifti_paths.append(path)

        additional_maps = self._add_new_maps(nifti_paths)

        map_names = copy.copy(self._controller.get_model().get_config().maps_to_show)
        map_names.extend(additional_maps)

        self._controller.apply_action(SetMapsToShow(map_names))

    def _save_settings(self):
        """Save the current settings as a text file.
        """
        current_model = self._controller.get_model()

        config_file = ['conf (*.conf)', 'All files (*)']
        file_name, used_filter = QFileDialog().getSaveFileName(caption='Select the GUI config file',
                                                               filter=';;'.join(config_file))
        if file_name:
            with open(file_name, 'w') as f:
                f.write(current_model.get_config().to_yaml())

    def _load_settings(self):
        config_file = ['conf (*.conf)', 'All files (*)']
        file_name, used_filter = QFileDialog().getOpenFileName(caption='Select the GUI config file',
                                                               filter=';;'.join(config_file))
        if file_name:
            with open(file_name, 'r') as f:
                try:
                    self._controller.apply_action(NewConfigAction(MapPlotConfig.from_yaml(f.read())))
                except yaml.parser.ParserError:
                    pass
                except yaml.scanner.ScannerError:
                    pass
                except ValueError:
                    pass

    def set_window_title(self, title):
        if title is None:
            self.setWindowTitle('MDT Maps Visualizer')
        else:
            self.setWindowTitle('MDT Maps Visualizer - {}'.format(title))


class PlottingFrameInfoToStatusBar(PlottingFrameInfoViewer):

    def __init__(self, controller, status_bar_label):
        super().__init__()
        self._controller = controller
        self._status_bar_label = status_bar_label

    def set_voxel_info(self, map_name, onscreen_coords, data_index):
        super().set_voxel_info(map_name, onscreen_coords, data_index)

        def format_value(v):
            value_format = '{:.3e}'
            if 1e-3 < v < 1e3:
                value_format = '{:.3f}'
            return value_format.format(v)

        data = self._controller.get_model().get_data()

        if map_name in data.get_map_names():
            value = data.get_map_data(map_name)[tuple(data_index)]
            clipped = value

            config = self._controller.get_model().get_config()
            if map_name in config.map_plot_options:
                clipped = config.map_plot_options[map_name].clipping.apply(value)

            if clipped != value:
                self._status_bar_label.setText("{}, {}, {} ({})".format(
                    onscreen_coords, data_index, format_value(clipped), format_value(value)))
            else:
                self._status_bar_label.setText("{}, {}, {}".format(onscreen_coords, data_index, format_value(value)))

    def clear_voxel_info(self):
        super().clear_voxel_info()
        self._status_bar_label.setText("")


class ExportImageDialog(Ui_SaveImageDialog, QDialog):

    previous_values = {'width': None, 'height': None,
                       'dpi': None, 'output_file': None,
                       'writeScriptsAndConfig': False}

    def __init__(self, parent, plotting_frame, controller):
        super().__init__(parent)
        self._extension_filters = [['png', '(*.png)'], ['svg', '(*.svg)']]
        self.setupUi(self)
        self._plotting_frame = plotting_frame
        self._controller = controller

        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.buttonBox.button(QDialogButtonBox.Ok).clicked.connect(self._export_image)
        self.outputFile_box.textChanged.connect(self._update_ok_button)
        self.outputFile_chooser.clicked.connect(lambda: self._select_file())

        if self.previous_values['width']:
            self.width_box.setValue(self.previous_values['width'])
        else:
            self.width_box.setValue(self._plotting_frame.width())

        if self.previous_values['height']:
            self.height_box.setValue(self.previous_values['height'])
        else:
            self.height_box.setValue(self._plotting_frame.height())

        if self.previous_values['dpi']:
            self.dpi_box.setValue(self.previous_values['dpi'])
        if self.previous_values['output_file']:
            self.outputFile_box.setText(self.previous_values['output_file'])
        if self.previous_values['writeScriptsAndConfig'] is not None:
            self.writeScriptsAndConfig.setChecked(self.previous_values['writeScriptsAndConfig'])

    @staticmethod
    def plot_frame_resized():
        ExportImageDialog.previous_values['width'] = None
        ExportImageDialog.previous_values['height'] = None

    @pyqtSlot()
    def _update_ok_button(self):
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(self.outputFile_box.text() != '')

    def _select_file(self):
        graphical_image_filters = [' '.join(el) for el in self._extension_filters] + ['All files (*)']

        open_file, used_filter = QFileDialog().getSaveFileName(caption='Select the output file',
                                                               filter=';;'.join(graphical_image_filters))

        if not any(open_file.endswith(el[0]) for el in self._extension_filters):
            extension_from_filter = list(filter(lambda v: ' '.join(v) == used_filter, self._extension_filters))
            if extension_from_filter:
                extension = extension_from_filter[0][0]
            else:
                extension = self._extension_filters[0][0]

            open_file += '.{}'.format(extension)

        if open_file:
            self.outputFile_box.setText(open_file)
            self._update_ok_button()

    def _export_image(self):
        output_file = self.outputFile_box.text()
        if not any(output_file.endswith(el[0]) for el in self._extension_filters):
            output_file += '.{}'.format(self._extension_filters[0][0])

        try:
            self._plotting_frame.export_image(output_file, self.width_box.value(), self.height_box.value(),
                                              dpi=self.dpi_box.value())
            self.previous_values['width'] = self.width_box.value()
            self.previous_values['height'] = self.height_box.value()
            self.previous_values['dpi'] = self.dpi_box.value()
            self.previous_values['output_file'] = self.outputFile_box.text()
            self.previous_values['writeScriptsAndConfig'] = self.writeScriptsAndConfig.isChecked()

            if self.writeScriptsAndConfig.isChecked():
                output_basename = os.path.splitext(output_file)[0]
                self._write_config_file(output_basename + '.conf')
                self._write_python_script_file(output_basename + '_script.py', output_basename + '.conf', output_file,
                                             self.width_box.value(), self.height_box.value(), self.dpi_box.value())
                self._write_bash_script_file(output_basename + '_script.sh', output_basename + '.conf', output_file,
                                             self.width_box.value(), self.height_box.value(), self.dpi_box.value())

        except PermissionError as error:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Could not write the file to the given destination.")
            msg.setInformativeText(str(error))
            msg.setWindowTitle("Permission denied")
            msg.exec_()

    def _write_config_file(self, output_basename):
        current_model = self._controller.get_model()
        with open(output_basename, 'w') as f:
            f.write(current_model.get_config().to_yaml())

    def _write_python_script_file(self, script_fname, configuration_fname, output_image_fname, width, height, dpi):
        with open(script_fname, 'w') as f:
            f.write('#!/usr/bin/env python\n')
            f.write(dedent('''
                {header}

                import mdt

                with open({config!r}, 'r') as f:
                    config = f.read()

                mdt.view_maps(
                    {paths},
                    config=config,
                    save_filename={output_name!r},
                    figure_options={{'width': {width}, 'height': {height}, 'dpi': {dpi}}})

            ''').format(header=get_script_file_header_text({'Purpose': 'Generate a results figure'}),
                        paths='[' + ', '.join(['{el!r}'.format(el=el) for el in self._get_file_paths()]) + ']',
                        config=configuration_fname,
                        output_name=output_image_fname,
                        width=width, height=height, dpi=dpi))

    def _write_bash_script_file(self, script_fname, configuration_fname, output_image_fname, width, height, dpi):
        with open(script_fname, 'w') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write(dedent('''
                {header}

                mdt-view-maps \\
                    {paths} \\
                    --config "{config}" \\
                    --to-file "{output_name}" \\
                    --width {width} \\
                    --height {height} \\
                    --dpi {dpi}
            ''').format(header=get_script_file_header_text({'Purpose': 'Generate a results figure'}),
                        paths=' '.join(['{el!r}'.format(el=el) for el in self._get_file_paths()]),
                        config=configuration_fname,
                        output_name=output_image_fname,
                        width=width, height=height, dpi=dpi))

    def _get_file_paths(self):
        data = self._controller.get_model().get_data()
        return list(data.get_file_paths().values())


class AboutDialog(Ui_AboutDialog, QDialog):

    def __init__(self, parent):
        super().__init__(parent)
        self.setupUi(self)
        self.contentLabel.setText(self.contentLabel.text().replace('{version}', mdt.__version__))


def start_gui(data=None, config=None, controller=None, app_exec=True, show_maximized=False, window_title=None):
    """Start the GUI with the given data and configuration.

    Args:
        data (DataInfo): the initial set of data
        config (MapPlotConfig): the initial configuration
        controller (mdt.gui.maps_visualizer.base.QtController): the controller to use in the application
        app_exec (boolean): if true we execute the Qt application, set to false to disable.
        show_maximized (true): if we want to show the window in a maximized state
        window_title (str): the title of the window

    Returns:
        MapsVisualizerWindow: the generated window
    """
    controller = controller or QtController()

    enable_pyqt_exception_hook()
    app = QtManager.get_qt_application_instance()

    # catches the sigint
    timer = QTimer()
    timer.start(500)
    timer.timeout.connect(lambda: None)

    main = MapsVisualizerWindow(controller)
    main.set_window_title(window_title)
    signal.signal(signal.SIGINT, main.send_sigint)

    center_window(main)

    if show_maximized:
        main.showMaximized()

    main.show()

    if data is None:
        data = SimpleDataInfo({})
    controller.apply_action(NewDataAction(data, config=config), store_in_history=False)

    QtManager.add_window(main)
    if app_exec:
        QtManager.exec_()

    return main



import copy
from mdt.gui.maps_visualizer.base import SimpleDataConfigModel
from mdt.visualization.maps.base import Zoom, Point2d, SingleMapConfig


class ModelUpdateAction(object):

    def apply(self, data_config_model):
        """Apply the changes to the given model and return a new model.

        Args:
            data_config_model (mdt.gui.maps_visualizer.base.DataConfigModel): the current data and configuration

        Returns:
            mdt.gui.maps_visualizer.base.DataConfigModel: the updated/new model
        """
        raise NotImplementedError()

    def unapply(self):
        """Return the configuration as it was before the application of this function.

        Returns:
            mdt.gui.maps_visualizer.base.DataConfigModel: the previous model
        """
        raise NotImplementedError()


class NewDataAction(ModelUpdateAction):

    def __init__(self, data, config=None):
        """Sets the new data and (optional) configuration when applied.

        This class will change some parts of the configuration to make sure that removing or adding maps does not
        result in an improper configuration.
        """
        self._previous_model = None
        self._new_config = config
        self._new_data = data

    def apply(self, data_config_model):
        """Apply the changes to the given model and return a new model.

        Args:
            data_config_model (mdt.gui.maps_visualizer.base.DataConfigModel): the current data and configuration

        Returns:
            mdt.gui.maps_visualizer.base.DataConfigModel: the updated/new model
        """
        self._previous_model = data_config_model

        old_config = self._new_config or self._previous_model.get_config()
        valid_config = old_config.create_valid(self._new_data)

        return SimpleDataConfigModel(self._new_data, valid_config)

    def unapply(self):
        """Return the configuration as it was before the application of this function.

        Returns:
            mdt.gui.maps_visualizer.base.DataConfigModel: the previous model
        """
        return self._previous_model


class ConfigAction(ModelUpdateAction):

    def __init__(self):
        """An abstract implementation of a model update action meant for updating only the configuration."""
        self._previous_model = None
        self._previous_config = None

    def apply(self, data_config_model):
        """Apply the changes to the configuration and keeps the data intact.

        By default this method calls _apply(data_config_model) to facilitate quick implementation.
        """
        self._previous_model = data_config_model
        self._previous_config = data_config_model.get_config()

        new_config = copy.deepcopy(self._previous_config)
        updated_new_config = self._apply(new_config)

        if updated_new_config:
            return SimpleDataConfigModel(data_config_model.get_data(), updated_new_config)
        return SimpleDataConfigModel(data_config_model.get_data(), new_config)

    def unapply(self):
        """Return the configuration as it was before the application of this function.

        Returns:
            mdt.gui.maps_visualizer.base.DataConfigModel: the previous model
        """
        return self._previous_model

    def _apply(self, config):
        """Facilitates quick implementation, called by apply()

        One can set configuration changes immediately to the configuration in the given model. If None is returned we
        do nothing.

        Args:
            config (mdt.visualization.maps.base.MapPlotConfig): the current data and configuration

        Returns:
            mdt.visualization.maps.base.MapPlotConfig: the updated/new configuration
        """
        raise NotImplementedError()


class SimpleConfigAction(ConfigAction):

    config_attribute = None

    def __init__(self, new_value):
        """A simple configuration action this sets the given value to the config attribute of the configuration.

        The config_attribute can be a list, if so, we iteratively look up the corresponding attributes and change
        the last attribute element.
        """
        super().__init__()
        self.new_value = new_value

    def _apply(self, configuration):
        if isinstance(self.config_attribute, (list, tuple)):
            obj = None
            for element in self.config_attribute[:-1]:
                obj = getattr(configuration, element)
            setattr(obj, self.config_attribute[-1], self.new_value)
        else:
            setattr(configuration, self.config_attribute, self.new_value)
        return self._extra_actions(configuration)

    def _extra_actions(self, configuration):
        """Called by the default configuration action to apply additional changes"""
        return configuration


class SimpleMapSpecificConfigAction(SimpleConfigAction):

    config_attribute = None

    def __init__(self, map_name, new_value):
        super().__init__(new_value)
        self.map_name = map_name

    def _apply(self, configuration):
        if self.map_name not in configuration.map_plot_options:
            configuration.map_plot_options[self.map_name] = SingleMapConfig()

        single_map_config = super()._apply(
            configuration.map_plot_options[self.map_name])

        if single_map_config is None:
            del configuration.map_plot_options[self.map_name]

        return configuration

    def _extra_actions(self, configuration):
        single_map_config = super()._extra_actions(configuration)
        if single_map_config == SingleMapConfig():
            return None
        return single_map_config


class NewConfigAction(ConfigAction):

    def __init__(self, new_config):
        super().__init__()
        self.new_config = new_config

    def apply(self, data_config_model):
        self._previous_model = data_config_model
        return SimpleDataConfigModel(data_config_model.get_data(), self.new_config)

    def unapply(self):
        return self._previous_model


class SetDimension(SimpleConfigAction):

    config_attribute = 'dimension'

    def _extra_actions(self, configuration):
        if self.new_value != self._previous_config.dimension:
            data = self._previous_model.get_data()

            max_slice = data.get_max_slice_index(self.new_value, configuration.maps_to_show)
            if configuration.slice_index > max_slice:
                configuration = SetSliceIndex(max_slice // 2).apply(
                    SimpleDataConfigModel(data, configuration)).get_config()

            return SetZoom(Zoom.no_zoom()).apply(SimpleDataConfigModel(data, configuration)).get_config()


class SetSliceIndex(SimpleConfigAction):

    config_attribute = 'slice_index'


class SetVolumeIndex(SimpleConfigAction):

    config_attribute = 'volume_index'


class SetMapsToShow(SimpleConfigAction):

    config_attribute = 'maps_to_show'


class SetColormap(SimpleConfigAction):

    config_attribute = 'colormap'


class SetRotate(SimpleConfigAction):

    config_attribute = 'rotate'

    def _extra_actions(self, configuration):
        if self.new_value != self._previous_config.rotate:

            data = self._previous_model.get_data()

            new_rotation = self.new_value - self._previous_config.rotate
            if self._previous_config.flipud:
                new_rotation *= -1

            new_zoom = self._previous_config.zoom.get_rotated(
                new_rotation,
                data.get_max_x_index(configuration.dimension, self._previous_config.rotate,
                                     configuration.maps_to_show) + 1,
                data.get_max_y_index(configuration.dimension, self._previous_config.rotate,
                                     configuration.maps_to_show) + 1)

            return SetZoom(new_zoom).apply(SimpleDataConfigModel(data, configuration)).get_config()


class SetZoom(SimpleConfigAction):

    config_attribute = 'zoom'


class SetGeneralMask(SimpleConfigAction):

    config_attribute = 'mask_name'


class SetPlotTitle(SimpleConfigAction):

    config_attribute = 'title'


class SetMapTitle(SimpleMapSpecificConfigAction):

    config_attribute = 'title'


class SetMapColorbarLabel(SimpleMapSpecificConfigAction):

    config_attribute = 'colorbar_label'


class SetMapScale(SimpleMapSpecificConfigAction):

    config_attribute = 'scale'


class SetMapClipping(SimpleMapSpecificConfigAction):

    config_attribute = 'clipping'


class SetMapColormap(SimpleMapSpecificConfigAction):

    config_attribute = 'colormap'


class SetFont(SimpleConfigAction):

    config_attribute = 'font'


class SetShowAxis(SimpleConfigAction):

    config_attribute = 'show_axis'


class SetShowPlotColorbars(SimpleConfigAction):

    config_attribute = ['colorbar_settings', 'visible']


class SetShowPlotTitles(SimpleConfigAction):

    config_attribute = 'show_titles'


class SetColorbarLocation(SimpleConfigAction):

    config_attribute = ['colorbar_settings', 'location']


class SetColorBarNmrTicks(SimpleConfigAction):

    config_attribute = ['colorbar_settings', 'nmr_ticks']


class SetInterpolation(SimpleConfigAction):

    config_attribute = 'interpolation'


class SetFlipud(SimpleConfigAction):

    config_attribute = 'flipud'

    def _extra_actions(self, configuration):
        if self.new_value != self._previous_config.flipud:
            data = self._previous_model.get_data()

            max_y = data.get_max_y_index(configuration.dimension, configuration.rotate,
                                         configuration.maps_to_show) + 1

            new_p0_y = max_y - configuration.zoom.p1.y
            if new_p0_y >= max_y - 1:
                new_p0_y = 0

            new_p1_y = max_y - configuration.zoom.p0.y
            if new_p1_y >= max_y - 1:
                new_p1_y = max_y - 1

            new_zoom = Zoom(Point2d(configuration.zoom.p0.x, new_p0_y),
                            Point2d(configuration.zoom.p1.x, new_p1_y))

            return SetZoom(new_zoom).apply(SimpleDataConfigModel(data, configuration)).get_config()


class SetAnnotations(SimpleConfigAction):

    config_attribute = 'annotations'



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'MainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MapsVisualizer(object):
    def setupUi(self, MapsVisualizer):
        MapsVisualizer.setObjectName("MapsVisualizer")
        MapsVisualizer.resize(1000, 754)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/main/logo"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MapsVisualizer.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(MapsVisualizer)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame_2 = QtWidgets.QFrame(self.centralwidget)
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame_2)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.splitter = QtWidgets.QSplitter(self.frame_2)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(-1, -1, -1, 6)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.commandTabs = QtWidgets.QTabWidget(self.verticalLayoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.commandTabs.sizePolicy().hasHeightForWidth())
        self.commandTabs.setSizePolicy(sizePolicy)
        self.commandTabs.setMinimumSize(QtCore.QSize(250, 0))
        self.commandTabs.setObjectName("commandTabs")
        self.generalOptions = QtWidgets.QWidget()
        self.generalOptions.setAccessibleName("")
        self.generalOptions.setObjectName("generalOptions")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.generalOptions)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setSpacing(0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.generalTabPosition = QtWidgets.QGridLayout()
        self.generalTabPosition.setSpacing(0)
        self.generalTabPosition.setObjectName("generalTabPosition")
        self.gridLayout_4.addLayout(self.generalTabPosition, 0, 0, 1, 1)
        self.commandTabs.addTab(self.generalOptions, "")
        self.mapOptions = QtWidgets.QWidget()
        self.mapOptions.setObjectName("mapOptions")
        self.gridLayout = QtWidgets.QGridLayout(self.mapOptions)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.mapSpecificTabPosition = QtWidgets.QGridLayout()
        self.mapSpecificTabPosition.setSpacing(0)
        self.mapSpecificTabPosition.setObjectName("mapSpecificTabPosition")
        self.gridLayout.addLayout(self.mapSpecificTabPosition, 0, 0, 1, 1)
        self.commandTabs.addTab(self.mapOptions, "")
        self.textInfoTab = QtWidgets.QWidget()
        self.textInfoTab.setObjectName("textInfoTab")
        self.gridLayout_9 = QtWidgets.QGridLayout(self.textInfoTab)
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.textInfoTabPosition = QtWidgets.QGridLayout()
        self.textInfoTabPosition.setSpacing(0)
        self.textInfoTabPosition.setObjectName("textInfoTabPosition")
        self.gridLayout_9.addLayout(self.textInfoTabPosition, 0, 0, 1, 1)
        self.commandTabs.addTab(self.textInfoTab, "")
        self.verticalLayout.addWidget(self.commandTabs)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 6, -1, 0)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.auto_rendering = QtWidgets.QCheckBox(self.verticalLayoutWidget)
        self.auto_rendering.setObjectName("auto_rendering")
        self.horizontalLayout.addWidget(self.auto_rendering)
        self.manual_render = QtWidgets.QPushButton(self.verticalLayoutWidget)
        self.manual_render.setIconSize(QtCore.QSize(16, 16))
        self.manual_render.setObjectName("manual_render")
        self.horizontalLayout.addWidget(self.manual_render)
        self.undo_config = QtWidgets.QPushButton(self.verticalLayoutWidget)
        self.undo_config.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/main/arrow_undo.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.undo_config.setIcon(icon1)
        self.undo_config.setObjectName("undo_config")
        self.horizontalLayout.addWidget(self.undo_config)
        self.redo_config = QtWidgets.QPushButton(self.verticalLayoutWidget)
        self.redo_config.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/main/arrow_redo.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.redo_config.setIcon(icon2)
        self.redo_config.setObjectName("redo_config")
        self.horizontalLayout.addWidget(self.redo_config)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout.setStretch(0, 1)
        self.gridLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget.setObjectName("gridLayoutWidget")
        self.plotLayout = QtWidgets.QVBoxLayout(self.gridLayoutWidget)
        self.plotLayout.setContentsMargins(-1, -1, -1, 0)
        self.plotLayout.setSpacing(0)
        self.plotLayout.setObjectName("plotLayout")
        self.gridLayout_2.addWidget(self.splitter, 0, 0, 1, 1)
        self.verticalLayout_2.addWidget(self.frame_2)
        self.verticalLayout_2.setStretch(0, 1)
        MapsVisualizer.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MapsVisualizer)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1000, 27))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuAbout = QtWidgets.QMenu(self.menubar)
        self.menuAbout.setObjectName("menuAbout")
        MapsVisualizer.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MapsVisualizer)
        self.statusbar.setObjectName("statusbar")
        MapsVisualizer.setStatusBar(self.statusbar)
        self.actionQuit = QtWidgets.QAction(MapsVisualizer)
        self.actionQuit.setObjectName("actionQuit")
        self.actionSaveImage = QtWidgets.QAction(MapsVisualizer)
        self.actionSaveImage.setObjectName("actionSaveImage")
        self.actionAbout = QtWidgets.QAction(MapsVisualizer)
        self.actionAbout.setObjectName("actionAbout")
        self.actionExtra_plot_options = QtWidgets.QAction(MapsVisualizer)
        self.actionExtra_plot_options.setObjectName("actionExtra_plot_options")
        self.actionSave_settings = QtWidgets.QAction(MapsVisualizer)
        self.actionSave_settings.setObjectName("actionSave_settings")
        self.actionLoad_settings = QtWidgets.QAction(MapsVisualizer)
        self.actionLoad_settings.setObjectName("actionLoad_settings")
        self.actionAdd_new_files = QtWidgets.QAction(MapsVisualizer)
        self.actionAdd_new_files.setObjectName("actionAdd_new_files")
        self.action_Clear = QtWidgets.QAction(MapsVisualizer)
        self.action_Clear.setObjectName("action_Clear")
        self.actionNew_window = QtWidgets.QAction(MapsVisualizer)
        self.actionNew_window.setObjectName("actionNew_window")
        self.menuFile.addAction(self.actionAdd_new_files)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave_settings)
        self.menuFile.addAction(self.actionLoad_settings)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSaveImage)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionNew_window)
        self.menuFile.addAction(self.action_Clear)
        self.menuFile.addAction(self.actionQuit)
        self.menuAbout.addAction(self.actionAbout)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuAbout.menuAction())

        self.retranslateUi(MapsVisualizer)
        self.commandTabs.setCurrentIndex(0)
        self.actionQuit.triggered.connect(MapsVisualizer.close)
        QtCore.QMetaObject.connectSlotsByName(MapsVisualizer)

    def retranslateUi(self, MapsVisualizer):
        _translate = QtCore.QCoreApplication.translate
        MapsVisualizer.setWindowTitle(_translate("MapsVisualizer", "MDT Maps Visualizer"))
        self.commandTabs.setTabText(self.commandTabs.indexOf(self.generalOptions), _translate("MapsVisualizer", "General"))
        self.commandTabs.setTabText(self.commandTabs.indexOf(self.mapOptions), _translate("MapsVisualizer", "Maps"))
        self.commandTabs.setTabText(self.commandTabs.indexOf(self.textInfoTab), _translate("MapsVisualizer", "Textual"))
        self.auto_rendering.setText(_translate("MapsVisualizer", "Auto redraw"))
        self.manual_render.setToolTip(_translate("MapsVisualizer", "Manually redraw the figure"))
        self.manual_render.setText(_translate("MapsVisualizer", "Redraw"))
        self.undo_config.setToolTip(_translate("MapsVisualizer", "Undo"))
        self.redo_config.setToolTip(_translate("MapsVisualizer", "Redo"))
        self.menuFile.setTitle(_translate("MapsVisualizer", "&File"))
        self.menuAbout.setTitle(_translate("MapsVisualizer", "&Help"))
        self.actionQuit.setText(_translate("MapsVisualizer", "&Quit"))
        self.actionQuit.setShortcut(_translate("MapsVisualizer", "Ctrl+Q"))
        self.actionSaveImage.setText(_translate("MapsVisualizer", "&Save image"))
        self.actionSaveImage.setShortcut(_translate("MapsVisualizer", "Ctrl+S"))
        self.actionAbout.setText(_translate("MapsVisualizer", "&About"))
        self.actionExtra_plot_options.setText(_translate("MapsVisualizer", "&Extra plot options"))
        self.actionSave_settings.setText(_translate("MapsVisualizer", "&Export settings"))
        self.actionLoad_settings.setText(_translate("MapsVisualizer", "&Import settings"))
        self.actionAdd_new_files.setText(_translate("MapsVisualizer", "&Add new file(s)"))
        self.action_Clear.setText(_translate("MapsVisualizer", "&Clear"))
        self.actionNew_window.setText(_translate("MapsVisualizer", "&New window"))

from . import main_rc



import glob
import os
from PyQt5.QtCore import QObject, pyqtSignal

from mdt.gui.utils import UpdateDescriptor

__author__ = 'Robbert Harms'
__date__ = "2015-08-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SharedState(QObject):

    state_updated_signal = pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        """The shared state for the model fitting GUI

        Attributes:
            base_dir (str): the base dir for all file opening operations
            dimension_index (int): the dimension index used in various operations
            slice_index (int): the slice index used in various operations
        """
        super().__init__(*args, **kwargs)

        shared_attributes = {'base_dir': None,
                             'dimension_index': 0,
                             'slice_index': 0,
                             'output_folder': None}

        for key, value in shared_attributes.items():
            def get_attribute_setter(attribute_key):
                def setter(value):
                    setattr(self, attribute_key, value)
                return setter

            setattr(self, '_' + key, value)
            setattr(SharedState, key, UpdateDescriptor(key))
            setattr(self, 'set_' + key, get_attribute_setter(key))


def results_preselection_names(data):
    """Generate a list of useful map names to display.

    This is primarily to be used as argument to the config option ``maps_to_show`` in the function :func:`view_maps`.

    Args:
        data (str or dict or list of str): either a directory or a dictionary of results or a list of map names.

    Returns:
        list of str: the list of useful/filtered map names.
    """
    keys = []
    if isinstance(data, str):
        for extension in ('.nii', '.nii.gz'):
            for f in glob.glob(os.path.join(data, '*' + extension)):
                keys.append(os.path.basename(f)[0:-len(extension)])
    elif isinstance(data, dict):
        keys = data.keys()
    else:
        keys = data

    filter_match = ('.vec', '.d', '.sigma', '.theta', '.phi', 'AIC', 'Errors', 'Errors', '.eigen_ranking',
                    'SignalEstimates', 'UsedMask', 'BIC')
    return list(sorted(filter(lambda v: all(m not in v for m in filter_match), keys)))




__author__ = 'Robbert Harms'
__date__ = "2015-10-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRIOptimizable(object):

    def __init__(self, *args, **kwargs):
        """This is an interface for some base methods we expect in an MRI model.

        Since we have both composite dMRI models and cascade models we must have an overarching interface to make
        sure that both type of models implement the same additional methods.

        The methods in this interface have little to do with modelling, but unify some extra required methods
        in the cascades and composite models.
        """
        super().__init__()

    def is_input_data_sufficient(self, input_data=None):
        """Check if the input data has enough information for this model to work.

        Args:
            input_data (mdt.utils.MRIInputData): The input data we intend on using with this model.

        Returns:
            boolean: True if there is enough information in the input data, false otherwise.
        """

    def get_input_data_problems(self, input_data=None):
        """Get all the problems with the protocol.

        Args:
            input_data (mdt.utils.MRIInputData): The input data we intend on using with this model.

        Returns:
            list of InputDataProblem: A list of
                InputDataProblem instances or subclasses of that baseclass.
                These objects indicate the problems with the protocol and this model.
        """

    def get_required_protocol_names(self):
        """Get a list with the constant data names that are needed for this model to work.

        For example, an implementing diffusion MRI model might require the presence of the protocol parameter
        ``g`` and ``b``. This function should then return ``('g', 'b')``.

        Returns:
            :class:`list`: A list of columns names that are to be taken from the protocol data.
        """

    def update_active_post_processing(self, processing_type, settings):
        """Update the active post-processing semaphores.

        It is possible to control which post-processing routines get run by overwriting them using this method.
        For a list of post-processors, please see the default mdt configuration file under ``active_post_processing``.

        Args:
            processing_type (str): one of ``sample`` or ``optimization``.
            settings (dict): the items to set in the post-processing information
        """
        raise NotImplementedError()


class InputDataProblem(object):

    def __init__(self):
        """The base class for indicating problems with the input data.

        These are meant to be returned from the function get_input_data_problems().

        Each of these problems is supposed to overwrite the function __str__() for reporting the problem.
        """

    def __repr__(self):
        return self.__str__()


class MissingProtocolInput(InputDataProblem):

    def __init__(self, missing_columns):
        super().__init__()
        self.missing_columns = missing_columns

    def __str__(self):
        return 'Missing columns: ' + ', '.join(self.missing_columns)


class NamedProtocolProblem(InputDataProblem):

    def __init__(self, model_protocol_problem, model_name):
        """This extends the given model protocol problem to also include the name of the model.

        Args:
            model_protocol_problem (InputDataProblem): The name for the problem with the given model.
            model_name (str): the name of the model
        """
        super().__init__()
        self._model_protocol_problem = model_protocol_problem
        self._model_name = model_name

    def __str__(self):
        return "{0}: {1}".format(self._model_name, self._model_protocol_problem)



__author__ = 'Robbert Harms'
__date__ = '2017-06-19'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class PlottingFrame(object):

    def __init__(self, controller, plotting_info_viewer=None):
        super().__init__()
        self._controller = controller
        self._plotting_info_viewer = plotting_info_viewer or NoOptPlottingFrameInfoViewer()

    def set_auto_rendering(self, auto_render):
        """Set if this plotting frame should auto render itself on every configuration update, or not.

        Args:
            auto_render (boolean): if True the plotting frame should auto render, if False it should only
                render on manual updates.
        """

    def redraw(self):
        """Tell the plotting frame to do a redraw."""

    def export_image(self, filename, width, height, dpi=100):
        """Export the current view as an image.

        Args:
            filename (str): where to write the file
            width (int): the width in pixels
            height (int): the height in pixels
            dpi (int): the dpi of the result
        """


class PlottingFrameInfoViewer(object):

    def __init__(self):
        """Implementations of this class can be given to a PlottingFrame to update viewing information.

        As an interface is bridges the gap between the rest of the GUI and the PlottingFrame and
        can encapsulate highlighting interesting aspects of one of the plots.
        """

    def set_voxel_info(self, map_name, onscreen_coords, data_index):
        """Highlight a single voxel.

        Args:
            map_name (str): the name of the map for which we are displaying the value
            onscreen_coords (tuple of x,y): the coordinates of the voxel onscreen
            data_index (tuple of x,y,z,v): the 4d coordinates of the corresponding voxel in the data
        """

    def clear_voxel_info(self):
        """Tell the info viewer that we are no longer looking at a specific voxel."""


class NoOptPlottingFrameInfoViewer(PlottingFrameInfoViewer):

    def set_voxel_info(self, map_name, onscreen_coords, data_index):
        super().set_voxel_info(map_name, onscreen_coords, data_index)

    def clear_voxel_info(self):
        super().clear_voxel_info()



import numpy as np
from mot.lib.cl_data_type import SimpleCLDataType
from mot.lib.cl_function import CLFunction, SimpleCLFunction


__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterPrior(CLFunction):
    """The priors are used during model sample, indicating the a priori information one has about a parameter.

    These priors are not in log space, we take the log in the model builder.

    The signature of prior parameters must be:

    .. code-block: c

            mot_float_type <prior_fname>(mot_float_type value,
                                         mot_float_type lower_bound,
                                         mot_float_type upper_bound,
                                         <extra_parameters>)
    """


class SimplePrior(ParameterPrior, SimpleCLFunction):

    def __init__(self, prior_name, prior_body, extra_params=None, dependencies=()):
        """A prior template function.

        This will prepend to the given extra parameters the obligatory parameters (value, lower_bound, upper_bound).

        Args:
            prior_name (str): the name of this prior function
            prior_body (str): the body of the prior
            extra_params (list): additional parameters for this prior
            dependencies (list or tuple): the list of dependency functions
        """
        extra_params = extra_params or []
        parameters = [('mot_float_type', 'value'),
                      ('mot_float_type', 'lower_bound'),
                      ('mot_float_type', 'upper_bound')] + extra_params
        super().__init__('mot_float_type', prior_name, parameters, prior_body,
                                          dependencies=dependencies)


class AlwaysOne(SimplePrior):

    def __init__(self):
        """The uniform prior is always 1. :math:`P(v) = 1` """
        super().__init__('uniform', 'return 1;')


class ReciprocalPrior(SimplePrior):

    def __init__(self):
        """The reciprocal of the current value. :math:`P(v) = 1/v` """
        body = '''
            if(value <= 0){
                return 0;
            }
            return 1.0/value;
        '''
        super().__init__('reciprocal', body)


class UniformWithinBoundsPrior(SimplePrior):

    def __init__(self):
        """This prior is 1 within the upper and lower bound of the parameter, 0 outside."""
        super().__init__(
            'uniform_within_bounds',
            'return value >= lower_bound && value <= upper_bound;')


class AbsSinPrior(SimplePrior):

    def __init__(self):
        """Angular prior: :math:`P(v) = |\\sin(v)|`"""
        super().__init__('abs_sin', 'return fabs(sin(value));')


class AbsSinHalfPrior(SimplePrior):

    def __init__(self):
        """Angular prior: :math:`P(v) = |\\sin(x)/2.0|`"""
        super().__init__('abs_sin_half', 'return fabs(sin(value)/2.0);')


class VagueGammaPrior(SimplePrior):

    def __init__(self):
        """The vague gamma prior is meant as a proper uniform prior.

        Lee & Wagenmakers:

            The practice of assigning Gamma(0.001, 0.001) priors on precision parameters is theoretically motivated by
            scale invariance arguments, meaning that priors are chosen so that changing the measurement
            scale of the data does not affect inference.
            The invariant prior on precision lambda corresponds to a uniform distribution on log sigma,
            that is, rho(sigma^2) prop. to. 1/sigma^2, or a Gamma(a -> 0, b -> 0) distribution.
            This invariant prior distribution, however, is improper (i.e., the area under the curve is unbounded),
            which means it is not really a distribution, but the limit of a sequence of continuous_distributions
            (see Jaynes, 2003). WinBUGS requires the use of proper continuous_distributions,
            and the Gamma(0.001, 0.001) prior is intended as a proper approximation to the theoretically
            motivated improper prior. This raises the issue of whether inference is sensitive to the essentially
            arbitrary value 0.001, and it is sometimes the case that using other small values such as 0.01 or 0.1
            leads to more stable sample
            in WinBUGS.

            -- Lee & Wagenmakers, Bayesian Cognitive Modeling, 2014, Chapter 4, Box 4.1

        While this is not WinBUGS and improper priors are allowed in MOT, it is still useful to have this prior
        in case people desire proper priors.
        """
        body = '''
            float kappa = 0.001;
            float theta = 1/0.001;

            return (1.0 / (tgamma(kappa) * pow(theta, kappa))) * pow(value, kappa - 1) * exp(- value / theta);
        '''
        super().__init__('vague_gamma_prior', body)


class NormalPDF(SimplePrior):

    def __init__(self):
        r"""Normal PDF on the given value: :math:`P(v) = N(v; \mu, \sigma)`"""
        from mdt.model_building.parameters import FreeParameter
        extra_params = [FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'mu', True, 0, -np.inf, np.inf,
                                      sampling_prior=AlwaysOne()),
                        FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'sigma', True, 1, -np.inf, np.inf,
                                      sampling_prior=AlwaysOne())]

        super().__init__(
            'normal_pdf',
            'return exp((mot_float_type) (-((value - mu) * (value - mu)) / (2 * sigma * sigma))) '
            '           / (sigma * sqrt(2 * M_PI));',
            extra_params)


class AxialNormalPDF(SimplePrior):

    def __init__(self):
        r"""The axial normal PDF is a Normal distribution wrapped around 0 and :math:`\pi`.

        It's PDF is given by:

        .. math::

            f(\theta; a, b) = \frac{\cosh(a\sin \theta + b\cos \theta)}{\pi I_{0}(\sqrt{a^{2} + b^{2}})}

        where in this implementation :math:`a` and :math:`b` are parameterized with the input variables
        :math:`\mu` and :math:`\sigma` using:

        .. math::

            \begin{align*}
            \kappa &= \frac{1}{\sigma^{2}} \\
            a &= \kappa * \sin \mu \\
            b &= \kappa * \cos \mu
            \end{align*}

        References:
            Barry C. Arnold, Ashis SenGupta (2006). Probability continuous_distributions and statistical inference for axial data.
            Environmental and Ecological Statistics, volume 13, issue 3, pages 271-285.
        """
        from mdt.model_building.parameters import FreeParameter
        from mot.library_functions import LogCosh, LogBesseli0

        extra_params = [FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'mu', True, 0, -np.inf, np.inf,
                                      sampling_prior=AlwaysOne()),
                        FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'sigma', True, 1, -np.inf, np.inf,
                                      sampling_prior=AlwaysOne())]

        super().__init__(
            'axial_normal_pdf',
            '''
                float kappa = 1.0 / pown(sigma, 2);
                float a = kappa * sin(mu);
                float b = kappa * cos(mu);

                return exp(log_cosh(a * sin(value) + b * cos(value))
                            - log_bessel_i0(sqrt(pown(a, 2) + pown(b, 2)))
                            - log(M_PI) );
            ''',
            extra_params,
            dependencies=(LogBesseli0(), LogCosh()))


class ARDBeta(SimplePrior):

    def __init__(self):
        r"""This is a collapsed form of the Beta PDF meant for use in Automatic Relevance Detection sample.

        In this prior the ``alpha`` parameter of the Beta prior is set to 1 which simplifies the equation.
        The parameter ``beta`` is still free and can be changed as desired.

        The implemented prior is:

        .. math::

            B(x; 1, \beta) = \beta * (1 - x)^{\beta - 1}

        """
        from mdt.model_building.parameters import FreeParameter
        extra_params = [FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'beta', False, 1, 1e-4, 1000,
                                      sampling_prior=ReciprocalPrior(),
                                      sampling_proposal_std=0.01)]

        body = '''
            if(value < 0 || value > 1){
                return 0;
            }
            return beta * pow(1 - value, beta - 1);
        '''
        super().__init__('ard_beta_pdf', body, extra_params)


class ARDGaussian(SimplePrior):

    def __init__(self):
        """This is a Gaussian prior meant for use in Automatic Relevance Detection sample.

        This uses a Gaussian prior with mean at zero and a standard deviation determined by the ``alpha`` parameter
        with the relationship :math:`\sigma = 1/\\sqrt(\\alpha)`.
        """
        from mdt.model_building.parameters import FreeParameter
        extra_params = [FreeParameter(SimpleCLDataType.from_string('mot_float_type'), 'alpha', False, 8, 1e-5, 1e4,
                                      sampling_prior=UniformWithinBoundsPrior(),
                                      sampling_proposal_std=20)]

        body = '''
            if(value < 0 || value > 1){
                return 0;
            }
            mot_float_type sigma = 1.0/sqrt(alpha);
            return exp(-pown(value, 2) / (2 * pown(sigma, 2))) / (sigma * sqrt(2 * M_PI));
        '''
        super().__init__('ard_beta_pdf', body, extra_params)



from mdt import CascadeTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI(CascadeTemplate):

    description = 'Cascade for NODDI initialized from Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'NODDI')
    inits = {'NODDI': {'w_ic.w': lambda output_previous, _: output_previous['w_stick0.w'] / 2.0,
                       'w_ec.w': lambda output_previous, _: output_previous['w_stick0.w'] / 2.0,
                       'w_csf.w': 'w_ball.w',
                       'NODDI_IC.theta': 'Stick0.theta',
                       'NODDI_IC.phi': 'Stick0.phi'}}


class NODDIDA(CascadeTemplate):

    description = 'Cascade for NODDIDA initialized from Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'NODDIDA')
    inits = {'NODDIDA': {'NODDI_IC.theta': 'Stick0.theta',
                         'NODDI_IC.phi': 'Stick0.phi'}}


class NODDI_Fixed(CascadeTemplate):

    cascade_name_modifier = 'fixed'
    description = 'Cascade for NODDI with fixed directions from Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'NODDI')
    inits = {'NODDI': {'w_ic.w': lambda output_previous, _: output_previous['w_stick0.w'] / 2.0,
                       'w_ec.w': lambda output_previous, _: output_previous['w_stick0.w'] / 2.0,
                       'w_csf.w': 'w_ball.w'}}
    fixes = {'NODDI': {'NODDI_IC.theta': 'Stick0.theta',
                       'NODDI_IC.phi': 'Stick0.phi'}}



import numpy as np
from matplotlib.ticker import LinearLocator

__author__ = 'Robbert Harms'
__date__ = "2014-02-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MyColourBarTickLocator(LinearLocator):

    def __init__(self, min_val, max_val, round_precision=3, **kwargs):
        super().__init__(**kwargs)
        self.min_val = min_val
        self.max_val = max_val
        self.round_precision = round_precision

    def __call__(self):
        locations = LinearLocator.__call__(self)

        new_locations = []
        for location in locations:
            if np.absolute(location) < 0.01:
                new_locations.append(float("{:.1e}".format(location)))
            else:
                new_locations.append(np.round(location, self.round_precision))

        if np.isclose(new_locations[-1], self.max_val) or new_locations[-1] >= self.max_val:
            new_locations[-1] = self.max_val

        if new_locations[0] <= self.min_val:
            new_locations[0] = self.min_val

        return new_locations



import logging
import numpy as np
from scipy.ndimage import binary_dilation, generate_binary_structure, binary_fill_holes
from mdt.utils import load_brain_mask
from mdt.protocols import load_protocol
from mdt.lib.nifti import load_nifti, write_nifti
import mot.configuration
from scipy.ndimage.filters import median_filter

__author__ = 'Robbert Harms'
__date__ = "2015-07-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def create_median_otsu_brain_mask(dwi_info, protocol, mask_threshold=0, fill_holes=True, **kwargs):
    """Create a brain mask using the given volume.

    Args:
        dwi_info (string or tuple or image): The information about the volume, either:

            - the filename of the input file
            - or a tuple with as first index a ndarray with the DWI and as second index the header
            - or only the image as an ndarray
        protocol (string or :class:`~mdt.protocols.Protocol`): The filename of the protocol file or a Protocol object
        mask_threshold (float): everything below this b-value threshold is masked away (value in s/m^2)
        fill_holes (boolean): if we will fill holes after the median otsu algorithm and before the thresholding
        **kwargs: the additional arguments for median_otsu.

    Returns:
        ndarray: The created brain mask
    """
    logger = logging.getLogger(__name__)
    logger.info('Starting calculating a brain mask')

    if isinstance(dwi_info, str):
        signal_img = load_nifti(dwi_info)
        dwi = signal_img.get_data()
    elif isinstance(dwi_info, (tuple, list)):
        dwi = dwi_info[0]
    else:
        dwi = dwi_info

    if isinstance(protocol, str):
        protocol = load_protocol(protocol)

    if len(dwi.shape) == 4:
        unweighted_ind = protocol.get_unweighted_indices()
        if len(unweighted_ind):
            unweighted = np.mean(dwi[..., unweighted_ind], axis=3)
        else:
            unweighted = np.mean(dwi, axis=3)
    else:
        unweighted = dwi.copy()

    brain_mask = median_otsu(unweighted, **kwargs)
    brain_mask = brain_mask > 0

    if fill_holes:
        brain_mask = binary_fill_holes(brain_mask)

    if mask_threshold:
        brain_mask = np.mean(dwi[..., protocol.get_weighted_indices()], axis=3) * brain_mask > mask_threshold

    logger.info('Finished calculating a brain mask')

    return brain_mask


def generate_simple_wm_mask(fa_fname, brain_mask_fname, out_fname, fa_threshold=0.3, median_radius=1, numpass=2):
    """Generate a simple white matter mask by thresholding the given FA map.

    Everything below the given FA threshold will be masked (not used). It also applies the regular brain mask to
    only retain values inside the brain.

    Args:
        fa_fname (str): the path to the FA file
        brain_mask_fname (str): the path to the general brain mask in use
        out_fname (str): where to write the outfile.
        fa_threshold (double): the FA threshold. Everything below this threshold is masked (set to 0). To be precise:
            where fa_data < fa_threshold set the value to 0.
        median_radius (int): the radius of the median filter
        numpass (int): the number of passes we apply the median filter
    """
    logger = logging.getLogger(__name__)
    logger.info('Starting calculating a white matter mask using FA.')

    nib_container = load_nifti(fa_fname)
    fa_data = nib_container.get_data()

    fa_data[fa_data < fa_threshold] = 0
    fa_data[fa_data > 0] = 1

    if len(fa_data.shape) > 3:
        fa_data = fa_data[:, :, :, 0]

    filter_footprint = np.zeros((1 + 2 * median_radius,) * 3)
    filter_footprint[median_radius, median_radius, median_radius] = 1
    filter_footprint[:, median_radius, median_radius] = 1
    filter_footprint[median_radius, :, median_radius] = 1
    filter_footprint[median_radius, median_radius, :] = 1

    mask = load_brain_mask(brain_mask_fname)

    fa_data_masked = np.ma.masked_array(fa_data, mask=mask)
    for ind in range(numpass):
        fa_data_masked = median_filter(fa_data_masked, footprint=filter_footprint, mode='constant')

    write_nifti(fa_data_masked, out_fname, nib_container.get_header())
    logger.info('Finished calculating a white matter mask.')


def create_write_median_otsu_brain_mask(dwi_info, protocol, output_fname, **kwargs):
    """Write a brain mask using the given volume and output as the given volume.

    Args:
        dwi_info (string or tuple or ndarray): the filename of the input file or a tuple with as
            first index a ndarray with the DWI and as second index the header or only the image.
        protocol (string or :class:`~mdt.protocols.Protocol`): The filename of the protocol file or a Protocol object
        output_fname (string): the filename of the output file (the extracted brain mask)
            If None, no output is written. If ``dwi_info`` is an ndarray also no file is written
            (we don't have the header).

    Returns:
        ndarray: The created brain mask
    """
    if isinstance(dwi_info, str):
        signal_img = load_nifti(dwi_info)
        dwi = signal_img.get_data()
        header = signal_img.get_header()
    else:
        dwi = dwi_info[0]
        header = dwi_info[1]

    mask = create_median_otsu_brain_mask(dwi, protocol, **kwargs)
    write_nifti(mask, output_fname, header)

    return mask


def median_otsu(unweighted_volume, median_radius=4, numpass=4, dilate=1):
    """ Simple brain extraction tool for dMRI data.

    This function is inspired from the ``median_otsu`` function from ``dipy``
    and is copied here to remove a dependency.

    It uses a median filter smoothing of the ``unweighted_volume``
    automatic histogram Otsu thresholding technique, hence the name
    *median_otsu*.

    This function is inspired from Mrtrix's bet which has default values
    ``median_radius=3``, ``numpass=2``. However, from tests on multiple 1.5T
    and 3T data. From GE, Philips, Siemens, the most robust choice is
    ``median_radius=4``, ``numpass=4``.

    Args:
        unweighted_volume (ndarray): ndarray of the unweighted volumes brain volumes
        median_radius (int): Radius (in voxels) of the applied median filter (default 4)
        numpass (int): Number of pass of the median filter (default 4)
        dilate (None or int): optional number of iterations for binary dilation

    Returns:
        ndarray: a 3D ndarray with the binary brain mask
    """
    b0vol = unweighted_volume

    logger = logging.getLogger(__name__)
    logger.info('We will use a single precision float type for the calculations.'.format())
    for env in mot.configuration.get_load_balancer().get_used_cl_environments(mot.configuration.get_cl_environments()):
        logger.info('Using device \'{}\'.'.format(str(env)))

    for ind in range(numpass):
        b0vol = median_filter(b0vol, size=median_radius, mode='mirror')

    thresh = _otsu(b0vol)
    mask = b0vol > thresh

    if dilate is not None:
        cross = generate_binary_structure(3, 1)
        mask = binary_dilation(mask, cross, iterations=dilate)

    return mask


def _otsu(image, nbins=256):
    """
    Return threshold value based on Otsu's method.
    Copied from scikit-image to remove dependency.

    Parameters
    ----------
    image : array
        Input image.
    nbins : int
        Number of bins used to calculate histogram. This value is ignored for
        integer arrays.

    Returns
    -------
    threshold : float
        Threshold value.
    """
    hist, bin_centers = np.histogram(image, nbins)
    hist = hist.astype(np.float)

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]

    # class means for all possible thresholds
    mean1 = np.cumsum(hist * bin_centers[1:]) / weight1
    mean2 = (np.cumsum((hist * bin_centers[1:])[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of `weight1`/`mean1` should pair with zero values in
    # `weight2`/`mean2`, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:])**2

    idx = np.argmax(variance12)
    threshold = bin_centers[:-1][idx]
    return threshold



from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mdt.lib.components import has_component, get_component
from mot.lib.cl_data_type import SimpleCLDataType
from mdt.model_building.parameter_functions.numdiff_info import NumDiffInfo, SimpleNumDiffInfo
from mdt.model_building.parameters import ProtocolParameter, FreeParameter
from mdt.model_building.parameter_functions.priors import UniformWithinBoundsPrior
from mdt.model_building.parameter_functions.transformations import AbstractTransformation


__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class DMRICompositeModel

        Args:
            template (Type[ParameterTemplate]): the configuration for the parameter.
        """
        data_type = template.data_type
        if isinstance(data_type, str):
            data_type = SimpleCLDataType.from_string(data_type)

        if issubclass(template, ProtocolParameterTemplate):
            class AutoProtocolParameter(ProtocolParameter):
                def __init__(self, nickname=None):
                    super().__init__(data_type, nickname or template.name, value=template.value)

            for name, method in template.bound_methods.items():
                setattr(AutoProtocolParameter, name, method)

            return AutoProtocolParameter

        elif issubclass(template, FreeParameterTemplate):
            numdiff_info = template.numdiff_info
            if not isinstance(numdiff_info, NumDiffInfo) and numdiff_info is not None:
                numdiff_info = SimpleNumDiffInfo(**numdiff_info)

            class AutoFreeParameter(FreeParameter):
                def __init__(self, nickname=None):
                    super().__init__(
                        data_type,
                        nickname or template.name,
                        template.fixed,
                        template.init_value,
                        template.lower_bound,
                        template.upper_bound,
                        parameter_transform=_resolve_parameter_transform(template.parameter_transform),
                        sampling_proposal_std=template.sampling_proposal_std,
                        sampling_prior=template.sampling_prior,
                        numdiff_info=numdiff_info
                    )
                    self.sampling_proposal_modulus = template.sampling_proposal_modulus

            for name, method in template.bound_methods.items():
                setattr(AutoFreeParameter, name, method)

            return AutoFreeParameter


class ParameterTemplate(ComponentTemplate):
    """The cascade template to inherit from.

    These templates are loaded on the fly by the ParametersBuilder

    template options:
        name (str): the name of the parameter, defaults to the class name
        description (str): the description of this parameter
        data_type (str or DataType): either a string we use as datatype or the actual datatype itself
    """
    _component_type = 'parameters'
    _builder = ParameterBuilder()

    name = ''
    description = ''
    data_type = 'mot_float_type'


class ProtocolParameterTemplate(ParameterTemplate):
    """The default template options for protocol parameters.
    """
    data_type = 'mot_float_type'
    value = None


class FreeParameterTemplate(ParameterTemplate):
    """The default template options for free parameters.

    Attributes:
        init_value (float): the initial value
        fixed (boolean or ndarray of float): if this parameter is fixed or not. If not fixed this should
            hold a reference to a value or a matrix
        lower_bound (float): the lower bounds
        upper_bound (float): the upper bounds
        parameter_transform
            (str or :class:`~mdt.model_building.parameter_functions.transformations.AbstractTransformation`): the
            parameter transformation, this is used for automatic range transformation of the parameters during
            optimization. See Harms 2017 NeuroImage for details. Typical elements are:

            * ``Identity``: no transformation
            * ``Positivity``: ensures the parameters are positive
            * ``Clamp``: limits the parameter between its lower and upper bounds
            * ``CosSqrClamp``: changes the range of the optimized parameters to [0, 1] and ensures boundary constraints
            * ``SinSqrClamp``: same as ``CosSqrClamp``
            * ``SqrClamp``: same as clamp but with an additional square root to change the magnitude of the range
            * ``AbsModPi``: ensures absolute modulus of the input parameters between zero and pi.
            * ``AbsModTwoPi``: ensures absolute modulus of the input parameters between zero and two pi.

        sampling_proposal_std (float): the default proposal standard deviation for this parameter. This is used
            in some MCMC sample routines.
        sampling_proposal_modulus (float or None): if given, a modulus we will use when finalizing the proposal
            continuous_distributions. That is, when we are finalizing the proposals we will take, if set, the absolute
            modulus of that parameter to ensure the parameter is within [0, <modulus>].
        sampling_prior: the prior function
        numdiff_info (dict or :class:`~mdt.model_building.parameter_functions.numdiff_info.NumDiffInfo`):
            the information necessary to take the numerical derivative of a model with respect to this parameter.
            Either a dictionary with the keyword arguments to
            :class:`~mdt.model_building.parameter_functions.numdiff_info.SimpleNumDiffInfo` or an information
            object directly. If None, we use an empty dictionary. Please note that if you override this, you will have
            to specify all of the items (no automatic inheritance of sub-items).
    """
    data_type = 'mot_float_type'
    fixed = False
    init_value = 0.03
    lower_bound = 0.0
    upper_bound = 4.0
    parameter_transform = 'Identity'
    sampling_proposal_std = 1
    sampling_proposal_modulus = None
    sampling_prior = UniformWithinBoundsPrior()
    numdiff_info = {'max_step': 0.1, 'scale_factor': 1, 'use_bounds': True,
                    'use_upper_bound': True, 'use_lower_bound': True}


def _resolve_parameter_transform(parameter_transform):
    """Resolves input parameter transforms to actual objects.

    Args:
        parameter_transform
            (str or :class:`~mdt.model_building.parameter_functions.transformations.AbstractTransformation`):
            a parameter transformation name (with or without the postfix ``Transform``) or an actual object we
            just return.

    Returns:
        mdt.model_building.parameter_functions.transformations.AbstractTransformation: an actual transformation object

    Raises:
        ValueError: if the parameter transformation could not be resolved.
    """
    if isinstance(parameter_transform, AbstractTransformation):
        return parameter_transform

    if has_component('parameter_transforms', parameter_transform):
        return get_component('parameter_transforms', parameter_transform)()

    if has_component('parameter_transforms', parameter_transform + 'Transform'):
        return get_component('parameter_transforms', parameter_transform + 'Transform')()

    raise ValueError('Could not resolve the parameter transformation "{}"'.format(parameter_transform))



"""Definitions of the protocol parameters.

The type of these parameters signifies that the data for this parameter should come from the protocol file or from
the protocol maps. These parameters are never optimized and are always set to the given input data.
"""
from mdt import ProtocolParameterTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class g(ProtocolParameterTemplate):
    data_type = 'mot_float_type4'


class gx(ProtocolParameterTemplate):
    pass


class gy(ProtocolParameterTemplate):
    pass


class gz(ProtocolParameterTemplate):
    pass


class b(ProtocolParameterTemplate):
    pass


class G(ProtocolParameterTemplate):
    pass


class Delta(ProtocolParameterTemplate):
    pass


class delta(ProtocolParameterTemplate):
    pass


class TE(ProtocolParameterTemplate):
    pass


class TM(ProtocolParameterTemplate):
    pass


class TI(ProtocolParameterTemplate):
    pass


class TR(ProtocolParameterTemplate):
    pass


class flip_angle(ProtocolParameterTemplate):
    value = 1


# For STEAM/TSE sequences, depending on the model in which they are used.
class Refoc_fa1(ProtocolParameterTemplate):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class Refoc_fa2(ProtocolParameterTemplate):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class SEf(ProtocolParameterTemplate):
    pass


class b1(ProtocolParameterTemplate):
    value = 1


class Sw(ProtocolParameterTemplate):
    """This parameter is created only for linear T1 decay fitting of GRE data with variable flip angle.

    S_weighted is defined as the input data divided by the
        :math:`tan(flip_angle) ->  S_weighted = data / tan (flip_angle * B1_map)`
    """
    value = 1


# For STEAM/TSE sequences, depending on the model in which they are used.
class Dt(ProtocolParameterTemplate):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class excitation_b1(ProtocolParameterTemplate):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class refocusing1_b1(ProtocolParameterTemplate):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class refocusing2_b1(ProtocolParameterTemplate):
    pass




__author__ = 'Robbert Harms'
__date__ = "2015-03-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tree(object):

    def __init__(self, data=None, tag=None, children=None, parent=None):
        """Create a new Tree.

        In this tree, every node is a tree object as well. The tree is implemented as a linked list.
        Each node has a reference to its children and to its parent node.

        Args:
            data (object): The data object
            tag (str): The tag used for displaying this node
            children (list of Tree): The list of children to this node
            parent (Tree): The parent tree node.

        Attributes:
            data (object): The data object
            tag (str): The tag used for displaying this node
            children (list of Tree): The list of children to this node
            parent (Tree): The parent tree node.
        """
        self.data = data
        self.tag = tag or ""
        self.children = children or []
        self.parent = parent

    @property
    def leaves(self):
        """Get all the leaves under this tree.

        Returns:
            list: A list of all leaves under this tree.
        """
        leaves = []
        if not self.children:
            leaves.append(self)
        else:
            for child in self.children:
                leaves.extend(child.leaves)
        return leaves

    @property
    def internal_nodes(self):
        """Get all the non-leaves under this tree (the internal nodes).

        Returns:
            list: A list of all non-leaves under this tree.
        """
        internal_nodes = []
        if self.children:
            internal_nodes.append(self)
            for child in self.children:
                internal_nodes.extend(child.internal_nodes)
        return internal_nodes

    def __str__(self, level=0):
        ret = "\t"*level + self.tag + "\n"
        for child in self.children:
            ret += child.__str__(level+1)
        return ret


class CompartmentModelTree(Tree):

    def __init__(self, model_lists):
        """Builds a multi modal multi compartment model from the given model tree.

        Valid model trees abides this grammar:

        tree     ::= model | '(' tree ')' | '(' tree ',' operator ')'
        model    ::= ModelCLFunction
        operator ::= '*' | '/' | '+' | '-'

        This means that one can build complex models consisting of "Model Functions" (for example,
        compartment models) that are combined using basic math operators.

        Args:
            model_lists (model tree list): The model tree list
        """
        super().__init__()
        self._init_tree(model_lists)

    def get_compartment_models(self):
        """Get the compartment models that are part of this tree.

        This basically just returns the leaves of the tree.

        Returns:
            list of ModelCLFunction: the compartments in this tree
        """
        return [n.data for n in self.leaves]

    def _init_tree(self, listing):
        if isinstance(listing, (list, tuple)):
            if len(listing) == 1:
                self.data = listing[0]
                self.tag = listing[0].name
            else:
                operator = None
                for node in listing:
                    if isinstance(node, str):
                        if operator is not None:
                            raise ValueError('Double operator in model listing.')
                        operator = node
                    else:
                        nn = CompartmentModelTree(node)
                        nn.parent = self
                        self.children.append(nn)

                if operator is None:
                    raise ValueError('No operator in model listing.')

                self.data = operator
                self.tag = operator
        else:
            self.data = listing
            self.tag = listing.name

    def __str__(self, level=0):
        if isinstance(self.data, str):
            operator = ' ' + self.data + ' '
            return '(' + "\n" + "\t" * (level + 1) + \
                   operator.join([child.__str__(level + 1) for child in self.children]) + \
                   "\n" + "\t" * level + ')'
        else:
            return self.data.name + '(' + ', '.join([p.name for p in self.data.get_parameters()]) + ')'



import numpy as np
import itertools

from mdt.lib.post_processing import DTIMeasures, DKIMeasures, noddi_dti_maps

from mdt.model_building.parameter_functions.priors import AlwaysOne, UniformWithinBoundsPrior
from mdt.component_templates.parameters import FreeParameterTemplate, ParameterBuilder
from mdt.component_templates.compartment_models import CompartmentTemplate
from mdt.model_building.parameter_functions.transformations import IdentityTransform, PositivityTransform

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def get_symmetric_indices(length, dimensions):
    """Get the indices we will use for indexing a symmetric matrix.

    This returns the indices of a lower triangular matrix.

    Args:
        length (int): the length of each side of the cube
        dimensions (int): the number of dimensions

    Returns:
        list of tuple: for every index numerical positions
    """
    def descending_or_equal(indices):
        """Predicate to decide if a index is part of the lower triangle of the matrix"""
        if len(indices) <= 1:
            return True
        if len(indices) == 2:
            return indices[0] >= indices[1]
        else:
            return (indices[0] >= indices[1]) and descending_or_equal(indices[1:])

    all_indices = list(itertools.product(*([range(length)] * dimensions)))
    return list(filter(descending_or_equal, all_indices))


def build_param(index):
    """Get one of the parameters of the Kurtosis matrix as an object.

    Args:
        tuple: index, 4 integers specyfing the location of this parameter in the matrix.

    Returns:
        Parameter: a constructed parameter to be used in a compartment model.
    """
    _lower_bound = -np.inf
    _parameter_transform = IdentityTransform()
    _sampling_prior = AlwaysOne()
    if len(set(index)) == 1:
        _lower_bound = 0
        _parameter_transform = PositivityTransform()
        _sampling_prior = UniformWithinBoundsPrior()

    class matrix_element_param(FreeParameterTemplate):
        name = 'W_{}{}{}{}'.format(*index)
        init_value = 0
        lower_bound = _lower_bound
        upper_bound = np.inf
        parameter_transform = _parameter_transform
        sampling_prior = _sampling_prior
        sampling_proposal_std = 0.01

    return ParameterBuilder().create_class(matrix_element_param)()


def get_parameters():
    """Get the list of parameters for the Kurtosis model.

    Returns:
        list: a list of parameters, some as a string some as actual parameters.
    """
    parameters = ['g', 'b', 'd', 'dperp0', 'dperp1', 'theta', 'phi', 'psi']

    for index in get_symmetric_indices(3, 4):
        parameters.append(build_param(index))

    return tuple(parameters)


class KurtosisTensor(CompartmentTemplate):

    description = '''
        The Kurtosis Tensor model.
    '''
    parameters = get_parameters()
    dependencies = ['TensorApparentDiffusion', 'KurtosisMultiplication']
    cl_code = '''
        mot_float_type adc = TensorApparentDiffusion(theta, phi, psi, d, dperp0, dperp1, g);

        if(adc <= 0.0){
            return 1;
        }

        mot_float_type tensor_md_2 = pown((d + dperp0 + dperp1) / 3.0, 2);

        double kurtosis_sum = KurtosisMultiplication(
            W_0000, W_1111, W_2222, W_1000, W_2000, W_1110,
            W_2220, W_2111, W_2221, W_1100, W_2200, W_2211,
            W_2100, W_2110, W_2210, g);

        if(kurtosis_sum < 0 || (((tensor_md_2 * b) / adc) * kurtosis_sum) > 3.0){
            return INFINITY;
        }

        return exp(-b*adc + (b*b)/6.0 * tensor_md_2 * kurtosis_sum);
    '''

    extra_prior = 'return dperp1 < dperp0 && dperp0 < d;'

    auto_add_cartesian_vector = False
    post_optimization_modifiers = [
        DTIMeasures.post_optimization_modifier
    ]
    extra_optimization_maps = [
        DTIMeasures.extra_optimization_maps,
        DKIMeasures.extra_optimization_maps,
        noddi_dti_maps
    ]
    extra_sampling_maps = [DTIMeasures.extra_sampling_maps]



from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array


__author__ = 'Robbert Harms'
__date__ = '2017-05-29'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


def wrap_objective_function(objective_function, decode_function, nmr_parameters):
    """Decorates the given objective function with parameter decoding.

    Args:
        objective_function (mot.lib.cl_function.CLFunction): A CL function with the signature:

            .. code-block:: c

                double <func_name>(local const mot_float_type* const x,
                                   void* data,
                                   local mot_float_type* objective_list);
        decode_function (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                transform the parameters from encoded space to model space so they can be used as input to the model.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);

        nmr_parameters (int): the number of parameters we are decoding.

    Returns:
        mot.lib.cl_function.CLFunction: the wrapped objective function.
    """
    return SimpleCLFunction.from_string('''
        double wrapped_''' + objective_function.get_cl_function_name() + '''(
                local const mot_float_type* const x,
                void* data, 
                local mot_float_type* objective_list){

            local mot_float_type x_model[''' + str(nmr_parameters) + '''];

            if(get_local_id(0) == 0){
                for(uint i = 0; i < ''' + str(nmr_parameters) + '''; i++){
                    x_model[i] = x[i];
                }
                ''' + decode_function.get_cl_function_name() + '''(data, x_model);
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            return ''' + objective_function.get_cl_function_name() + '''(
                x_model, data, objective_list);    
        }
    ''', dependencies=[objective_function, decode_function])


class ParameterCodec(object):

    def __init__(self, encode_func, decode_func):
        """Create a parameter codec container.

        Args:
            encode_func (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                transform the parameters from model space to encoded space so they can be used as input to an
                CL routine. The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);

            decode_func (mot.lib.cl_function.CLFunction): An OpenCL function that is used in the CL kernel to
                transform the parameters from encoded space to model space so they can be used as input to the model.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
        """
        self._encode_func = encode_func
        self._decode_func = decode_func

    def get_encode_function(self):
        """Get a CL function that can transform the model parameters from model space to an encoded space.

        Returns:
            mot.lib.cl_function.CLFunction: An OpenCL function that is used in the CL kernel to transform the parameters
                from model space to encoded space so they can be used as input to an CL routine.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
        """
        return self._encode_func

    def get_decode_function(self):
        """Get a CL function that can transform the model parameters from encoded space to model space.

        Returns:
            mot.lib.cl_function.CLFunction: An OpenCL function that is used in the CL kernel to transform the parameters
                from encoded space to model space so they can be used as input to the model.
                The signature of the CL function is:

                .. code-block:: c

                    void <fname>(void* data, local mot_float_type* x);
        """
        return self._decode_func

    def decode(self, parameters, kernel_data=None, cl_runtime_info=None):
        """Decode the given parameters using the given model.

        This transforms the data from optimization space to model space.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        return self._transform_parameters(self.get_decode_function(),
                                          parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    def encode(self, parameters, kernel_data=None, cl_runtime_info=None):
        """Encode the given parameters using the given model.

        This transforms the data from model space to optimization space.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        return self._transform_parameters(self.get_encode_function(),
                                          parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    def encode_decode(self, parameters, kernel_data, codec, cl_runtime_info=None):
        """First apply an encoding operation and then apply a decoding operation again.

        This can be used to enforce boundary conditions in the parameters.

        Args:
            parameters (ndarray): The parameters to transform
            kernel_data (dict[str: mot.lib.utils.KernelData]): the additional data to load
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

        Returns:
            ndarray: The array with the transformed parameters.
        """
        encode_func = codec.get_encode_function()
        decode_func = codec.get_decode_function()

        func = SimpleCLFunction.from_string('''
            void encode_decode_parameters(void* data, local mot_float_type* x){
                ''' + encode_func.get_cl_function_name() + '''(data, x);
                ''' + decode_func.get_cl_function_name() + '''(data, x);
            }
        ''', dependencies=[encode_func, decode_func])
        return self._transform_parameters(func, parameters, kernel_data, cl_runtime_info=cl_runtime_info)

    @staticmethod
    def _transform_parameters(cl_func, parameters, kernel_data, cl_runtime_info=None):
        cl_named_func = SimpleCLFunction.from_string('''
            void transformParameterSpace(void* data, local mot_float_type* x){
                ''' + cl_func.get_cl_function_name() + '''(data, x);
            }
        ''', dependencies=[cl_func])

        kernel_data = {'data': kernel_data,
                       'x': Array(parameters, ctype='mot_float_type', mode='rw')}

        cl_named_func.evaluate(kernel_data, parameters.shape[0], cl_runtime_info=cl_runtime_info)
        return kernel_data['x'].get_data()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Get the MDT example data that is accompanying the installation.

This will write the MDT example data (b1k_b2k and b6k datasets) to the indicated directory. You can use this data for
testing MDT on your computer. These example datasets are contained within the MDT package and as such are available
on every machine with MDT installed.
"""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
from mdt.lib.shell_utils import BasicShellApplication
import textwrap

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GetExampleData(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-get-example-data
            mdt-get-example-data .
        ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('dir', metavar='dir', type=str, nargs='?', help='the output directory',
                            default=None).completer = FilesCompleter()
        return parser

    def run(self, args, extra_args):
        if args.dir:
            output_dir = os.path.realpath(args.dir)
        else:
            output_dir = os.getcwd()

        mdt.get_example_data(output_dir)


def get_doc_arg_parser():
    return GetExampleData().get_documentation_arg_parser()


if __name__ == '__main__':
    GetExampleData().start()



from copy import deepcopy

__author__ = 'Robbert Harms'
__date__ = '2017-07-20'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


# The list of components we are loading at the moment. This allows keeping track of nested components.
_component_loading = []


class ComponentBuilder(object):
    """The base class for component builders.

    Component builders, together with ComponentTemplate allow you to define components using Templates,
    special classes where the properties are defined using class attributes.

    The ComponentTemplate contains class attributes defining the component which can be used by the
    ComponentBuilder to create a class of the right type from the information in that template.
    """

    def create_class(self, template):
        """Create a class of the right type given the information in the template.

        Args:
            template (ComponentTemplate): the information as a component config

        Returns:
            class: the class of the right type
        """
        from mdt.lib.components import temporary_component_updates, add_template_component

        cls = self._create_class(template)

        if template.subcomponents:
            subcomponents = template.subcomponents

            class SubComponentConstruct(cls):
                def __init__(self, *args, **kwargs):
                    with temporary_component_updates():
                        for component in subcomponents:
                            add_template_component(component)
                        super().__init__(*args, **kwargs)

            return SubComponentConstruct
        return cls

    def _create_class(self, template):
        """Create a class of the right type given the information in the template.

        This is to be used by subclasses.

        Args:
            template (ComponentTemplate): the information as a component config

        Returns:
            class: the class of the right type
        """
        raise NotImplementedError()
    

def bind_function(func):
    """This decorator is for methods in ComponentTemplates that we would like to bind to the constructed component.

    Example suppose you want to inherit or overwrite a function in the constructed model, then in your template/config
    you should define the function and add @bind_function to it as a decorator, like this:

    .. code-block:: python

        # the class we want to create
        class MyGoal(object):
            def test(self):
                print('test')

        # the template class from which we want to construct a new MyGoal, note the @bind_function
        class MyConfig(ComponentTemplate):
            @bind_function
            def test(self):
                super().test()
                print('test2')

    The component builder takes care to actually bind the new method to the final object.

    What this will do essentially is that it will add the property bind to the function. This should act as a
    flag indicating that that function should be bound.

    Args:
        func (python function): the function to bind to the build object
    """
    func._bind = True
    return func


class ComponentTemplateMeta(type):

    def __new__(mcs, name, bases, attributes):
        """A pre-processor for the components.

        On the moment this meta class does two things, first it adds all functions with the '_bind' property
        to the ``bound_methods`` dictionary for binding them later to the constructed class. Second, it sets the
        ``name`` attribute to the template class name if there is no ``name`` attribute defined.
        """
        result = super().__new__(mcs, name, bases, attributes)

        result.component_type = mcs._resolve_attribute(bases, attributes, '_component_type')
        result.bound_methods = mcs._get_bound_methods(bases, attributes)
        result.subcomponents = mcs._get_subcomponents(attributes)
        result.name = mcs._get_component_name_attribute(name, bases, attributes)

        if len(_component_loading) == 1:
            try:
                if result.component_type is not None and result.name:
                    from mdt.lib.components import add_template_component
                    add_template_component(result)
            except ValueError:
                pass

        _component_loading.pop()
        return result

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        _component_loading.append(name)
        return dict()

    @staticmethod
    def _get_component_name_attribute(name, bases, attributes):
        if 'name' in attributes:
            return attributes['name']
        return name

    @staticmethod
    def _get_bound_methods(bases, attributes):
        """Get all methods in the template that have the ``_bind`` attribute.

        This collects all methods that have the ``_bind`` attribute to a dictionary.

        Returns:
            dict: all the methods that need to be bound.
        """
        bound_methods = {value.__name__: value for value in attributes.values() if hasattr(value, '_bind')}
        for base in bases:
            if hasattr(base, 'bound_methods'):
                for key, value in base.bound_methods.items():
                    if key not in bound_methods:
                        bound_methods.update({key: value})
        return bound_methods

    @staticmethod
    def _get_subcomponents(attributes):
        """Get all the sub-components defined in this template.

        Returns:
            list: the defined sub-components.
        """
        return [value for value in attributes.values() if isinstance(value, ComponentTemplateMeta)]

    @staticmethod
    def _resolve_attribute(bases, attributes, attribute_name, base_predicate=None):
        """Search for the given attribute in the given attributes or in the attributes of the bases.

        Args:
            base_predicate (func): if given, a predicate that runs on the attribute of one of the bases to determine
                if we will return that one.

        Returns:
            The value for the attribute

        Raises:
            ValueError: if the attribute could not be found in the attribute or any of the bases
        """
        base_predicate = base_predicate or (lambda _: True)

        if attribute_name in attributes:
            return attributes[attribute_name]
        for base in bases:
            if hasattr(base, attribute_name) and base_predicate(getattr(base, attribute_name)):
                return getattr(base, attribute_name)
        raise ValueError('Attribute not found in this component config or its superclasses.')


class ComponentTemplate(object, metaclass=ComponentTemplateMeta):
    """The component configuration.

    By overriding the class attributes you can define complex configurations. The actual class distilled from these
    configurations are loaded by the builder referenced by ``_builder``.

    Attributes:
        _component_type (str): the component type of this template. Set to one of the valid template types.
        _builder (ComponentBuilder): the builder to use for constructing an object of the given template
        name (str): the name of the template
        description (str): a description of the object / template
    """
    _component_type = None
    _builder = None
    name = ''
    description = ''

    def __new__(cls, *args, **kwargs):
        """Instead of creating an instance of a Template, this will build the actual component.

        This allows one to build the model of a template by regular object initialization. For example, these two
        calls (a and b) are exactly the same::

            template = Template
            a = construct_component(template)
            b = template()
        """
        return cls._builder.create_class(cls)

    @classmethod
    def meta_info(cls):
        return {'name': cls.name,
                'description': cls.description,
                'template': deepcopy(cls)}



from mdt import CascadeTemplate

__author__ = 'Robbert Harms'
__date__ = '2017-08-26'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class S0T1_MI_EPI(CascadeTemplate):

    description = 'Cascade for S0T1_MI_EPI model.'
    models = ('S0',
              ('S0T1_MI_EPI', 'S0T1_MI_EPI_fixed_eff'),
              'S0T1_MI_EPI')
    fixes = {'S0T1_MI_EPI_fixed_eff': {'ExpT1DecIR.Efficiency': 1}}


class S0_T2(CascadeTemplate):

    description = 'Cascade for S0-T2 model.'
    models = ('S0',
              'S0-T2')



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'generate_brain_mask_tab.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GenerateBrainMaskTabContent(object):
    def setupUi(self, GenerateBrainMaskTabContent):
        GenerateBrainMaskTabContent.setObjectName("GenerateBrainMaskTabContent")
        GenerateBrainMaskTabContent.resize(827, 427)
        self.verticalLayout = QtWidgets.QVBoxLayout(GenerateBrainMaskTabContent)
        self.verticalLayout.setContentsMargins(-1, 11, -1, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.verticalLayout.addLayout(self.verticalLayout_2)
        self.line = QtWidgets.QFrame(GenerateBrainMaskTabContent)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setHorizontalSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.label_5 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 1, 2, 1, 1)
        self.label_12 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 5, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.selectImageButton = QtWidgets.QPushButton(GenerateBrainMaskTabContent)
        self.selectImageButton.setObjectName("selectImageButton")
        self.horizontalLayout_2.addWidget(self.selectImageButton)
        self.selectedImageText = QtWidgets.QLineEdit(GenerateBrainMaskTabContent)
        self.selectedImageText.setText("")
        self.selectedImageText.setObjectName("selectedImageText")
        self.horizontalLayout_2.addWidget(self.selectedImageText)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 2, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.medianRadiusInput = QtWidgets.QSpinBox(GenerateBrainMaskTabContent)
        self.medianRadiusInput.setMinimum(1)
        self.medianRadiusInput.setProperty("value", 4)
        self.medianRadiusInput.setObjectName("medianRadiusInput")
        self.horizontalLayout_4.addWidget(self.medianRadiusInput)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_4, 4, 1, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.selectOutputButton = QtWidgets.QPushButton(GenerateBrainMaskTabContent)
        self.selectOutputButton.setObjectName("selectOutputButton")
        self.horizontalLayout_3.addWidget(self.selectOutputButton)
        self.selectedOutputText = QtWidgets.QLineEdit(GenerateBrainMaskTabContent)
        self.selectedOutputText.setObjectName("selectedOutputText")
        self.horizontalLayout_3.addWidget(self.selectedOutputText)
        self.gridLayout.addLayout(self.horizontalLayout_3, 2, 1, 1, 1)
        self.label_8 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 2, 0, 1, 1)
        self.label_11 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 4, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.selectProtocolButton = QtWidgets.QPushButton(GenerateBrainMaskTabContent)
        self.selectProtocolButton.setObjectName("selectProtocolButton")
        self.horizontalLayout_5.addWidget(self.selectProtocolButton)
        self.selectedProtocolText = QtWidgets.QLineEdit(GenerateBrainMaskTabContent)
        self.selectedProtocolText.setObjectName("selectedProtocolText")
        self.horizontalLayout_5.addWidget(self.selectedProtocolText)
        self.gridLayout.addLayout(self.horizontalLayout_5, 1, 1, 1, 1)
        self.line_3 = QtWidgets.QFrame(GenerateBrainMaskTabContent)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout.addWidget(self.line_3, 3, 0, 1, 3)
        self.label_10 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 2, 2, 1, 1)
        self.label_6 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_3.setMinimumSize(QtCore.QSize(0, 0))
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.label_13 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        self.label_13.setObjectName("label_13")
        self.gridLayout.addWidget(self.label_13, 6, 0, 1, 1)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.numberOfPassesInput = QtWidgets.QSpinBox(GenerateBrainMaskTabContent)
        self.numberOfPassesInput.setMinimum(1)
        self.numberOfPassesInput.setProperty("value", 4)
        self.numberOfPassesInput.setObjectName("numberOfPassesInput")
        self.horizontalLayout_6.addWidget(self.numberOfPassesInput)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem1)
        self.gridLayout.addLayout(self.horizontalLayout_6, 5, 1, 1, 1)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.finalThresholdInput = QtWidgets.QDoubleSpinBox(GenerateBrainMaskTabContent)
        self.finalThresholdInput.setLocale(QtCore.QLocale(QtCore.QLocale.C, QtCore.QLocale.AnyCountry))
        self.finalThresholdInput.setPrefix("")
        self.finalThresholdInput.setSuffix("")
        self.finalThresholdInput.setMaximum(1000000.0)
        self.finalThresholdInput.setObjectName("finalThresholdInput")
        self.horizontalLayout_7.addWidget(self.finalThresholdInput)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem2)
        self.gridLayout.addLayout(self.horizontalLayout_7, 6, 1, 1, 1)
        self.label_14 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 4, 2, 1, 1)
        self.label_15 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_15.setFont(font)
        self.label_15.setObjectName("label_15")
        self.gridLayout.addWidget(self.label_15, 5, 2, 1, 1)
        self.label_16 = QtWidgets.QLabel(GenerateBrainMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_16.setFont(font)
        self.label_16.setObjectName("label_16")
        self.gridLayout.addWidget(self.label_16, 6, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line_2 = QtWidgets.QFrame(GenerateBrainMaskTabContent)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 6, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.generateButton = QtWidgets.QPushButton(GenerateBrainMaskTabContent)
        self.generateButton.setEnabled(False)
        self.generateButton.setObjectName("generateButton")
        self.horizontalLayout.addWidget(self.generateButton)
        self.viewButton = QtWidgets.QPushButton(GenerateBrainMaskTabContent)
        self.viewButton.setEnabled(False)
        self.viewButton.setObjectName("viewButton")
        self.horizontalLayout.addWidget(self.viewButton)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem4 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem4)

        self.retranslateUi(GenerateBrainMaskTabContent)
        QtCore.QMetaObject.connectSlotsByName(GenerateBrainMaskTabContent)
        GenerateBrainMaskTabContent.setTabOrder(self.selectImageButton, self.selectedImageText)
        GenerateBrainMaskTabContent.setTabOrder(self.selectedImageText, self.selectProtocolButton)
        GenerateBrainMaskTabContent.setTabOrder(self.selectProtocolButton, self.selectedProtocolText)
        GenerateBrainMaskTabContent.setTabOrder(self.selectedProtocolText, self.selectOutputButton)
        GenerateBrainMaskTabContent.setTabOrder(self.selectOutputButton, self.selectedOutputText)
        GenerateBrainMaskTabContent.setTabOrder(self.selectedOutputText, self.medianRadiusInput)
        GenerateBrainMaskTabContent.setTabOrder(self.medianRadiusInput, self.numberOfPassesInput)
        GenerateBrainMaskTabContent.setTabOrder(self.numberOfPassesInput, self.finalThresholdInput)
        GenerateBrainMaskTabContent.setTabOrder(self.finalThresholdInput, self.generateButton)
        GenerateBrainMaskTabContent.setTabOrder(self.generateButton, self.viewButton)

    def retranslateUi(self, GenerateBrainMaskTabContent):
        _translate = QtCore.QCoreApplication.translate
        GenerateBrainMaskTabContent.setWindowTitle(_translate("GenerateBrainMaskTabContent", "Form"))
        self.label.setText(_translate("GenerateBrainMaskTabContent", "Generate brian mask"))
        self.label_2.setText(_translate("GenerateBrainMaskTabContent", "Create a whole brain mask using the median-otsu algorithm."))
        self.label_5.setText(_translate("GenerateBrainMaskTabContent", "(To create one, please see the tab \"Generate protocol file\")"))
        self.label_12.setText(_translate("GenerateBrainMaskTabContent", "Number of passes:"))
        self.selectImageButton.setText(_translate("GenerateBrainMaskTabContent", "Browse"))
        self.label_4.setText(_translate("GenerateBrainMaskTabContent", "(Select the 4d diffusion weighted image)"))
        self.selectOutputButton.setText(_translate("GenerateBrainMaskTabContent", "Browse"))
        self.label_8.setText(_translate("GenerateBrainMaskTabContent", "Select output file:"))
        self.label_11.setText(_translate("GenerateBrainMaskTabContent", "Median radius:"))
        self.selectProtocolButton.setText(_translate("GenerateBrainMaskTabContent", "Browse"))
        self.label_10.setText(_translate("GenerateBrainMaskTabContent", "(Default is <volume_name>_mask.nii.gz)"))
        self.label_6.setText(_translate("GenerateBrainMaskTabContent", "Select 4d image:"))
        self.label_3.setText(_translate("GenerateBrainMaskTabContent", "Select protocol file:"))
        self.label_13.setText(_translate("GenerateBrainMaskTabContent", "Final threshold:"))
        self.label_14.setText(_translate("GenerateBrainMaskTabContent", "(Radius (in voxels) of the applied median filter)"))
        self.label_15.setText(_translate("GenerateBrainMaskTabContent", "(Number of median filter passes)"))
        self.label_16.setText(_translate("GenerateBrainMaskTabContent", "(Additional masking threshold as a signal intensity)"))
        self.generateButton.setText(_translate("GenerateBrainMaskTabContent", "Generate"))
        self.viewButton.setText(_translate("GenerateBrainMaskTabContent", "View mask"))




import glob
import os

from mdt.lib.nifti import nifti_filepath_resolution
from mdt.utils import split_image_path


def get_shortest_unique_names(paths):
    """Get the shortest unique map name between two or more nifti file paths.

    This function is useful when loading multiple maps like for example ``./foo.nii`` and ``./foo.nii.gz`` or
    ``../directory0/foo.nii`` and ``../directory1/foo.nii``. In all cases the map name (foo) is similar,
    but we want to be able to show them distinctly. This function tries to find the shortest unique map names
    for each of the maps.

    Example output:
    * [``./foo.nii``] -> ``foo``
    * [``./foo.nii``, ``./foo.nii.gz``] -> [``foo.nii``, ``foo.nii.gz``]
    * [``../directory0/foo.nii``, ``../directory1/foo.nii``] -> [``directory0/foo``, ``directory1/foo``]

    Args:
        paths (list of str): the paths to the different nifti files

    Returns:
        tuple: the map names for the given set of paths (in the same order)
    """
    if not len(paths):
        return []

    real_paths = []
    for path in paths:
        if os.path.exists(path):
            real_paths.append(path)
        else:
            real_paths.append('/' + path)

    paths = real_paths

    dirs, names, exts = zip(*map(split_image_path, paths))

    if len(set(paths)) == 1:
        return names

    new_names = []

    multiple_dirs = len(set(dirs)) > 1
    common_prefix = os.path.commonpath(dirs)

    def multiple_maps_in_same_dir(current_ind, current_directory, current_name):
        for ind, (directory, name, ext, full_path) in enumerate(zip(dirs, names, exts, paths)):
            if ind != current_ind and directory == current_directory and name == current_name:
                return True
        return False

    for ind, (directory, name, ext, full_path) in enumerate(zip(dirs, names, exts, paths)):
        if multiple_dirs:
            new_name = os.path.relpath(directory, common_prefix) + '/' + name
            if new_name.startswith('../') or new_name.startswith('..\\'):
                new_name = new_name[3:]
            if common_prefix == '/' or common_prefix == '\\':
                new_name = '/' + new_name
        else:
            new_name = name

        if multiple_maps_in_same_dir(ind, directory, name):
            new_name += ext

        new_names.append(new_name)

    return new_names


def find_all_nifti_files(paths):
    """Find the paths to the nifti files in the given paths.

    For directories we add all the nifti files from that directory, for file we try to resolve them to nifti files.

    Args:
        paths (list of str): the list of file paths

    Returns:
        list of str: the list of all nifti files (no more directories)
    """
    def find_niftis(paths, recurse=True):
        niftis = []
        for path in paths:
            if os.path.isdir(path):
                if recurse:
                    niftis.extend(find_niftis(glob.glob(path + '/*.nii*'), recurse=False))
            else:
                try:
                    niftis.append(nifti_filepath_resolution(path))
                except ValueError:
                    pass
        return niftis
    return find_niftis(paths)
